/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QTimer>

#include "machine_dc.h"

static const int						TIMER_INTERVAL(40);

TMachineDC::TMachineDC(void)
{
	d_geo_enabled = true;
	d_shift_out_enabled = false;
}

TMachineDC::~TMachineDC(void)
{
}

void TMachineDC::Process_Rx(
	const QByteArray					&new_data)
{
	QByteArray							byte_sequence;
	int									rx_line_index(-1);
	int									rx_byte_index(-1);
	int									index;
	QStringList							list;
	QString								text;
	QString								write_text;
	QString								line;
	QString								keyword;
	QString								arguments;
	TVector3							pos;
	TVector3							vec;
	TVector3							p1,p2;
	double								a,b;
	bool								x_error,y_error,z_error;
	bool								send_channel_2;
	int									arg1,arg2,arg3,arg4,arg5,arg6;
	TCommandExecutor::TCommand			command;
	
	d_rx_data += new_data;
	
	if(d_shift_out_enabled)
	{
		this->Send_Clean(new_data);
	}

	do
	{
		if(d_rx_data.size())
		{
			
#if DEBUG_MODE_LEVEL_0
			QByteArray 	control_bytes;
			int			control_byte_cntr;
			char		byte;
			
			for(control_byte_cntr = 0;control_byte_cntr < d_rx_data.size();++control_byte_cntr)
			{
				byte = d_rx_data[control_byte_cntr];
				
				if(byte < 0x20 &&
				   byte != '\r' &&
				   byte != '\n')
				{
					control_bytes.push_back(byte);
				}
			}
			
			if(control_bytes.size())
			{
				qWarning(QString("RECV: %1").arg(QString(control_bytes.toHex())).toLatin1());
			}
#endif
			
			rx_byte_index = d_rx_data.indexOf(char(0x02));
			
			if(!(rx_byte_index < 0))
			{
				d_rx_data.remove(rx_byte_index,1);
				
#if DEBUG_MODE_LEVEL_1
				qWarning(QString("RECV: CTRL_B (0x02)").toLatin1());
#endif
				d_command_executor.Abort();
				
				d_jog_timer->stop();
				d_control_timer->stop();
				
				this->Send_Delayed("READY\r\n",50);
			}
			
			rx_byte_index = d_rx_data.indexOf(char(0x03));
			
			if(!(rx_byte_index < 0))
			{
				d_rx_data.remove(rx_byte_index,1);
				
#if DEBUG_MODE_LEVEL_1
				qWarning(QString("RECV: CTRL_C (0x03)").toLatin1());
#endif
				d_command_executor.Abort();
				
				d_jog_timer->stop();
				d_control_timer->stop();
			}
			
			rx_byte_index = d_rx_data.indexOf(char(0x05));
			
			if(!(rx_byte_index < 0))
			{
				d_rx_data.remove(rx_byte_index,1);
				d_shift_out_enabled = true;
				
#if DEBUG_MODE_LEVEL_1
				qWarning(QString("RECV: CTRL_E (0x05)").toLatin1());
#endif
				d_command_executor.Abort();
				
				d_jog_timer->stop();
				d_control_timer->stop();
			}
			
			rx_byte_index = d_rx_data.indexOf(char(0x08));
			
			if(!(rx_byte_index < 0))
			{
				byte_sequence.resize(3);
				byte_sequence[0] = char(0x08);
				byte_sequence[1] = char(0x20);
				byte_sequence[2] = char(0x08);
				
				this->Send(byte_sequence);
				
				d_rx_data.remove(rx_byte_index,1);
				d_rx_data.chop(1);		// remove last byte
				
#if DEBUG_MODE_LEVEL_1
				qWarning(QString("RECV: CTRL_H (0x08) Backspace").toLatin1());
#endif
			}
			
			rx_byte_index = d_rx_data.indexOf(char(0x0e));
			
			if(!(rx_byte_index < 0))
			{
				d_rx_data.remove(rx_byte_index,1);
				d_shift_out_enabled = false;
				
#if DEBUG_MODE_LEVEL_1
				qWarning(QString("RECV: CTRL_N (0x0e)").toLatin1());
#endif
			}
			
			rx_byte_index = d_rx_data.indexOf(char(0x1a));
			
			if(!(rx_byte_index < 0))
			{
				d_rx_data.remove(rx_byte_index,1);
				d_shift_out_enabled = false;
				
#if DEBUG_MODE_LEVEL_1
				qWarning(QString("RECV: CTRL_Z (0x1a)").toLatin1());
#endif
				d_command_executor.Abort();
				
				d_jog_timer->stop();
				d_control_timer->stop();
				
				this->Send("READY\r\n");
			}
		}
	}while(!(rx_byte_index < 0) && d_rx_data.size());
	
	
	do
	{
		rx_line_index = d_rx_data.indexOf('\r');
		
		if(!(rx_line_index < 0))
		{
			line = QString::fromLatin1(d_rx_data.mid(0,rx_line_index)).simplified();
			d_rx_data.remove(0,rx_line_index + 1);
			
			// clean text
			while(line.length() && line[0].toLatin1() < 0x20)
			{
				line = line.mid(1);
			}
			
			emit Log_Rx(line);
			
			if(d_shift_out_enabled)
			{
#if DEBUG_MODE_LEVEL_1
				qWarning(QString("INFO:  Shift Out Enabled").toLatin1());
#endif
				this->Send("\r\n");
			}
			
			index = line.indexOf(' ');
			
			if(!(index < 0))
			{
				keyword = line.mid(0,index).toUpper();
				arguments = line.mid(index + 1);
			}
			else
			{
				keyword = line.toUpper();
				arguments = QString();
			}
			
			send_channel_2 = false;
			
			if(keyword.startsWith(QStringLiteral(">2>")))
			{
				keyword = keyword.mid(3);
				send_channel_2 = true;
			}
			
#if DEBUG_MODE_LEVEL_1
			qWarning(QString("RECV: %1").arg(line).toLatin1());
#endif
			
			if(keyword.compare(QStringLiteral("ACCEL")) == 0)
			{
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("AUTZER")) == 0)
			{
				if(d_estop_status)
				{
					this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x14000040 E-Stop Push Button Pressed\r\n",send_channel_2);
				}
				else
				{
					pos = d_command_executor.Current_Position();
					p1 = d_min_machine;
					p2 = d_max_machine;
					
					p1.z = p2.z;
					
					command.command_type = TCommandExecutor::COMMAND_MOVE;
					command.target = TVector3(pos.x,pos.y,p1.z + d_tool_offset.z);
					command.speed = d_move_speed;
					command.completion_text = QString();
					
					d_command_executor.Add_Command(command);
					
					command.command_type = TCommandExecutor::COMMAND_MOVE;
					command.target = p1 + d_tool_offset;
					command.speed = d_move_speed;
					
					if(send_channel_2)
					{
						command.completion_text = QStringLiteral("<2<%\r\n");
					}
					else
					{
						command.completion_text = QStringLiteral("%\r\n");
					}
					
					d_command_executor.Add_Command(command);
					
					if(!d_control_timer->isActive())
					{
						d_control_timer->start(TIMER_INTERVAL);
					}
				}
			}
			else if(keyword.compare(QStringLiteral("CODENUM")) == 0)
			{
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("CMHWPF")) == 0)
			{
				this->Send("CMHWPF            4,        99,         4\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("CMSWPF")) == 0)
			{
				/*
				 Argument 1               Argument 2              Argument 3              Argument 4              Argument 5              Argument 6
				 0 CMM_KNOWS_CMMTYP     	 0 CMM_KNOWS_ACCEL   	 0 CMM_KNOWS_SCADAT  	 0 CMM_KNOWS_CMMACSY  	 0 CMM_KNOWS_SCLTMP  	 0 CMM_KNOWS_VMABRT
				 1 CMM_KNOWS_AUTZER     	 1 CMM_KNOWS_MOVABS  	 1 CMM_KNOWS_SCADIR  	 1 CMM_KNOWS_OFSTCSY  	 1 CMM_KNOWS_WKPTCSY 	 1 CMM_KNOWS_STATUS
				 2 CMM_KNOWS_AUTORP     	 2 CMM_KNOWS_MOVDLT  	 2 CMM_KNOWS_SCAINT  	 2 CMM_KNOWS_REFRCSY  	 2 CMM_KNOWS_WKPPAR  	 2 CMM_KNOWS_CANCEL
				 3 CMM_KNOWS_INITCM     	 3 CMM_KNOWS_MOVPAR  	 3 CMM_KNOWS_SCAN    	 3 CMM_KNOWS_RFIXCSY  	 3 CMM_KNOWS_PRBTPA  	 3 CMM_KNOWS_DUMP
				 4 CMM_KNOWS_RESET      	 4 CMM_KNOWS_GETPOS  	 4 CMM_KNOWS_SCAPNT  	 4 CMM_KNOWS_ROFSCSY  	 4 CMM_KNOWS_READTP  	 4 CMM_KNOWS_DEVFUN
				 5 CMM_KNOWS_DISABLE    	 5 CMM_KNOWS_GETDVM  	 5 CMM_KNOWS_SCASTA  	 5 CMM_KNOWS_RPHICSY  	 5 CMM_KNOWS_TWARNING	 5 CMM_KNOWS_CADATA
				 6 CMM_KNOWS_ENABLE     	 6 CMM_KNOWS_RPTPOS  	 6 CMM_KNOWS_SCASTO  	 8 CMM_KNOWS_JOYSTCSY 	16 CMM_KNOWS_INIMAC  	 6 CMM_KNOWS_AUTCAL
				 7 CMM_KNOWS_SHOW       	 7 CMM_KNOWS_GETPOSM 	 7 CMM_KNOWS_HSSDAT  	 9 CMM_KNOWS_DISPCSY  	17 CMM_KNOWS_DFNMAC  	 7 CMM_KNOWS_PSTOVM
				 8 CMM_KNOWS_VERSION    	16 CMM_KNOWS_PRBRPT  	 8 CMM_KNOWS_HSSAXI  	10 CMM_KNOWS_DSPINF   	18 CMM_KNOWS_EDTMAC  	 8 CMM_KNOWS_DEFINE
				 9 CMM_KNOWS_CMHWPF     	17 CMM_KNOWS_PRBATT  	 9 CMM_KNOWS_HSSCIR  	11 CMM_KNOWS_DIMENS   	19 CMM_KNOWS_USEMAC  	 9 CMM_KNOWS_XEFILE
				 10 CMM_KNOWS_CMHWST     18 CMM_KNOWS_PRBLPA  	10 CMM_KNOWS_HSSCYL  	12 CMM_KNOWS_SETMSK   	20 CMM_KNOWS_MPRBFS  	10 CMM_KNOWS_SET
				 11 CMM_KNOWS_CMSWPF     19 CMM_KNOWS_PRBGPA  	16 CMM_KNOWS_CENDAT  	16 CMM_KNOWS_MAGACSY  	21 CMM_KNOWS_SETOFS  	11 CMM_KNOWS_HELP
				 16 CMM_KNOWS_BCKTRK     20 CMM_KNOWS_PRBHTY  	17 CMM_KNOWS_CENSTA  	17 CMM_KNOWS_NEXTPRB    22 CMM_KNOWS_QUALSPH    12 CMM_KNOWS_VMTEST
				 17 CMM_KNOWS_IBOPEN     21 CMM_KNOWS_PRBTYP  	18 CMM_KNOWS_CENTER  	18 CMM_KNOWS_SETMAGA    23 CMM_KNOWS_CALMAT     13 CMM_KNOWS_ASCEFC
				 18 CMM_KNOWS_MOVTST     22 CMM_KNOWS_PRBPIN  	21 CMM_KNOWS_DPS_DAT 	24 CMM_KNOWS_PARK       24 CMM_KNOWS_RTADJUST   14 CMM_KNOWS_VMFRMT
				 19 CMM_KNOWS_QSDEV      23 CMM_KNOWS_PROBE   	22 CMM_KNOWS_DPS_AXI 	25 CMM_KNOWS_ENDPARK    25 CMM_KNOWS_QUALRONT   15 CMM_KNOWS_VMEXIT
				 20 CMM_KNOWS_PRBDATA    24 CMM_KNOWS_PRBRDV  	23 CMM_KNOWS_DPS_CIR    26 CMM_KNOWS_SETMAGA0   26 CMM_KNOWS_QUALSTRT   16 CMM_KNOWS_SETSNSR
				 21 CMM_KNOWS_PRESSURE   25 CMM_KNOWS_USETOOL    24 CMM_KNOWS_DPS_CYL    27 CMM_KNOWS_ENABLELGYPROT 27 CMM_KNOWS_CLEARCBH 17 CMM_KNOWS_PRBXPT
				 22 CMM_KNOWS_PRBTST     26 CMM_KNOWS_ALIGNTOOL  25 CMM_KNOWS_DPS_CON    28 CMM_KNOWS_SENSORMEASRANGE 28 CMM_KNOWS_SETWDWE 19 CMM_KNOWS_FLY
				 23 CMM_KNOWS_READVU     27 CMM_KNOWS_PRBCOEF    26 CMM_KNOWS_DPS_PLA    29 CMM_KNOWS_LOWLEVELMATRIX 29 CMM_KNOWS_SHOWNCESCL 20 CMM_KNOWS_HSS-TAG
				 24 CMM_KNOWS_TESTCORR   28 CMM_KNOWS_XTDPRBRDV  27 CMM_KNOWS_CONSCAN    30 CMM_KNOWS_RACKERROR  30 CMM_KNOWS_MOVSPDPC    21 CMM_KNOWS_VHSS
				 25 CMM_KNOWS_NEWDOWN    29 CMM_KNOWS_SNSDATA    28 CMM_KNOWS_SHOWRTTYP                                                   22 CMM_KNOWS_SRCNUM
				 26 CMM_KNOWS_BASECOMP   30 CMM_KNOWS_PTBVCMMDATA 29 CMM_KNOWS_MAXJOGAUTORETRACTDIST                                      23 CMM_KNOWS_DATFORM
				 27 CMM_KNOWS_PROBECOMP                          30 CMM_KNOWS_SHOWFEEDRATE                                                24 CMM_KNOWS_VHSSSPEEDANDDENSITY
				 28 CMM_KNOWS_THD                                                                                                         25 CMM_KNOWS_USERINTERACTION
				 29 CMM_KNOWS_CPAR                                                                                                        26 CMM_REQUIRES_AUTHORISATION
				 27 CMM_KNOWS_TOPOCORR
				 28 CMM_FORCE_SENDPRBTYP
				 29 CMM_KNOWS_STOP
				 30 CMM_KNOWS_CONTLINES
				 */
				
				arg1 = 0b00000000000000000000111011111011;
				arg2 = 0b00000000111111110000000000011011;
				arg3 = 0b00000000000000000000000000000000;
				arg4 = 0b00000000000000000000001001000000;
				arg5 = 0b00000000000000000000000000010100;
				arg6 = 0b00000000000010000000010000000010;
				
				write_text = QString("CMSWPF %1, %2, %3, %4, %5, %6\r\n").arg(arg1,9).arg(arg2,9).arg(arg3,9).arg(arg4,9).arg(arg5,9).arg(arg6,9);
				
				this->Send(write_text.toLatin1(),send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("CMHWST")) == 0)
			{
				/*
				 Argument 1                      Argument 2                          Argument 3
				 0 IS_ZEROED                   	 0 CMM_PART_TEMPCOMP_ENABLED      	 0 CMM_RFIXCSY_ENABLED
				 1 VOLCOMP_ENABLED             	 1 CMM_PROBE_TEMPCOMP_ENABLED     	 1 CMM_ROFSCSY_ENABLED
				 2 PROBE_REF_POINT_SET         	 2 CMM_TEMPCOMP_ENABLED           	 2 CMM_RPHICSY_ENABLED
				 3 PROBE_CHANGER_CSY_LOADED    	 3 CMM_ROTARY_TABLE_MOVE_ENABLED  	 3 CMM_OFSTCSY_ENABLED
				 4 ROTARY_TABLE_CALIBRATED     	 4 CMM_RTT_COMP_ENABLED           	 4 CMM_REFRCSY_ENABLED
				 5 DEFAULT_PARAMS_MODIFIED     	 5 CMM_GEO_COMP_ENABLED           	 5 CMM_CMMACSY_ENABLED
				 6 READY_TO_MOVE                                                      6 CMM_JOYSTCSY_ENABLED
				 7 QDS_OPEN                                                           7 CMM_DISPCSY_ENABLED
				 8 QDS_READY                                                          8 CMM_SENDS_PROBING_DIRECTION
				 9 QDS_PROBE_ACTIVE                                                   9 CMM_MOVE_VOLCOMPED
				 10 QDS_RTCSY_FOUND                                                   10 CMM_JOGBOX_ENABLED
				 11 QDS_PCCSY_FOUND                                                   11 CMM_USES_LARGE_BUFFER
				 12 PRB_CHANGER_PORTS_VALID                                           12 CMM_PROBE_IS_IN_HEAD
				 13 MOTORS_POWERED_UP
				 */
				arg1 = 0b0010000001000111;
				arg2 = 0b0000000000100000;
				arg3 = 0b0000011000000000;
				
				write_text = QString("CMHWST %1, %2, %3\r\n").arg(arg1,9).arg(arg2,9).arg(arg3,9);
				
				this->Send(write_text.toLatin1(),send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("CMMTYP")) == 0)
			{
				this->Send("CMMTYP   MA   1234, FDC       V16.01, DMP 20 30 21    , 0\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("DISABLE")) == 0)
			{
				if(arguments.startsWith(QStringLiteral("GEO"),Qt::CaseInsensitive))
				{
					d_geo_enabled = false;
				}
				
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("ENABLE")) == 0)
			{
				if(arguments.startsWith(QStringLiteral("GEO"),Qt::CaseInsensitive))
				{
					d_geo_enabled = true;
				}
				
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("GETPOS")) == 0)
			{
				pos = d_command_executor.Current_Position();
				
				write_text = QString("POS %1, %2, %3, 0.00000, 0.00000, 0.00000, 0.00000\r\n").arg(pos.x,0,'f',5).arg(pos.y,0,'f',5).arg(pos.z,0,'f',5);
				
				this->Send(write_text.toLatin1(),send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("INITCM")) == 0)
			{
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("JOGPBCSY")) == 0)
			{
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("JOYSTCSY")) == 0)
			{
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("MOVPAR")) == 0)		// MOVPAR 500.000000,500.000000,500.000000
			{
				d_move_speed = this->Get_Move_Speed(arguments);
				
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("MOVABS")) == 0)		// MOVABS 500.000000,1000.000000,-600.000000
			{
				if(d_estop_status)
				{
					this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x14000040 E-Stop Push Button Pressed\r\n");
				}
				else
				{
					pos = this->Get_Move_Data(arguments);
					
					if(this->Is_Position_Inside_Volume(pos,&x_error,&y_error,&z_error))
					{
						command.command_type = TCommandExecutor::COMMAND_MOVE;
						command.target = pos;
						command.speed = d_move_speed;
						
						if(send_channel_2)
						{
							command.completion_text = QStringLiteral("<2<%\r\n");
						}
						else
						{
							command.completion_text = QStringLiteral("%\r\n");
						}
						
						d_command_executor.Add_Command(command);
						
						if(!d_control_timer->isActive())
						{
							d_control_timer->start(TIMER_INTERVAL);
						}
					}
					else
					{
						if(x_error)
						{
							emit Log_Text(QStringLiteral("ERROR CE ,S%1 ,TR_CE ,0x1400104d Position out of the machine stroke limit on X axis"));
							this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x1400104d Position out of the machine stroke limit on X axis\r\n");
							d_shift_out_enabled = false;
						}
						
						if(y_error)
						{
							emit Log_Text(QStringLiteral("ERROR CE ,S%1 ,TR_CE ,0x1400204d Position out of the machine stroke limit on Y axis"));
							this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x1400204d Position out of the machine stroke limit on Y axis\r\n");
							d_shift_out_enabled = false;
						}
						
						if(z_error)
						{
							emit Log_Text(QStringLiteral("ERROR CE ,S%1 ,TR_CE ,0x1400304d Position out of the machine stroke limit on Z axis"));
							this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x1400304d Position out of the machine stroke limit on Z axis\r\n");
							d_shift_out_enabled = false;
						}
						
						d_command_executor.Abort();
						
						d_jog_timer->stop();
						d_control_timer->stop();
					}
				}
			}
			else if(keyword.compare(QStringLiteral("MOVCIR")) == 0)		// MOVCIR 500.000000,1000.000000,-600.000000
			{
				if(d_estop_status)
				{
					this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x14000040 E-Stop Push Button Pressed\r\n");
				}
				else
				{
					pos = this->Get_Move_Data(arguments);
					
					if(this->Is_Position_Inside_Volume(pos,&x_error,&y_error,&z_error))
					{
						command.command_type = TCommandExecutor::COMMAND_MOVECIR;
						command.target = pos;
						command.speed = d_move_speed;
						
						if(send_channel_2)
						{
							command.completion_text = QStringLiteral("<2<%\r\n");
						}
						else
						{
							command.completion_text = QStringLiteral("%\r\n");
						}
						
						d_command_executor.Add_Command(command);
						
						if(!d_control_timer->isActive())
						{
							d_control_timer->start(TIMER_INTERVAL);
						}
					}
					else
					{
						if(x_error)
						{
							emit Log_Text(QStringLiteral("ERROR CE ,S%1 ,TR_CE ,0x1400104d Position out of the machine stroke limit on X axis"));
							this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x1400104d Position out of the machine stroke limit on X axis\r\n");
							d_shift_out_enabled = false;
						}
						
						if(y_error)
						{
							emit Log_Text(QStringLiteral("ERROR CE ,S%1 ,TR_CE ,0x1400204d Position out of the machine stroke limit on Y axis"));
							this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x1400204d Position out of the machine stroke limit on Y axis\r\n");
							d_shift_out_enabled = false;
						}
						
						if(z_error)
						{
							emit Log_Text(QStringLiteral("ERROR CE ,S%1 ,TR_CE ,0x1400304d Position out of the machine stroke limit on Z axis"));
							this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x1400304d Position out of the machine stroke limit on Z axis\r\n");
							d_shift_out_enabled = false;
						}
						
						d_command_executor.Abort();
						
						d_jog_timer->stop();
						d_control_timer->stop();
					}
				}
			}
			else if(keyword.compare(QStringLiteral("MOVCTR")) == 0)		// center position of MOVCIR.  MOVCTR 500.000000,1000.000000,-600.000000
			{
				if(d_estop_status)
				{
					this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x14000040 E-Stop Push Button Pressed\r\n");
				}
				else
				{
					pos = this->Get_Move_Data(arguments);
					
					if(this->Is_Position_Inside_Volume(pos,&x_error,&y_error,&z_error))
					{
						command.command_type = TCommandExecutor::COMMAND_DEFINE_MOVECIR_CENTER;
						command.target = pos;
						
						if(send_channel_2)
						{
							command.completion_text = QStringLiteral("<2<%\r\n");
						}
						else
						{
							command.completion_text = QStringLiteral("%\r\n");
						}
						
						d_command_executor.Add_Command(command);
						
						if(!d_control_timer->isActive())
						{
							d_control_timer->start(TIMER_INTERVAL);
						}
					}
					else
					{
						if(x_error)
						{
							emit Log_Text(QStringLiteral("ERROR CE ,S%1 ,TR_CE ,0x1400104d Position out of the machine stroke limit on X axis"));
							this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x1400104d Position out of the machine stroke limit on X axis\r\n");
							d_shift_out_enabled = false;
						}
						
						if(y_error)
						{
							emit Log_Text(QStringLiteral("ERROR CE ,S%1 ,TR_CE ,0x1400204d Position out of the machine stroke limit on Y axis"));
							this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x1400204d Position out of the machine stroke limit on Y axis\r\n");
							d_shift_out_enabled = false;
						}
						
						if(z_error)
						{
							emit Log_Text(QStringLiteral("ERROR CE ,S%1 ,TR_CE ,0x1400304d Position out of the machine stroke limit on Z axis"));
							this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x1400304d Position out of the machine stroke limit on Z axis\r\n");
							d_shift_out_enabled = false;
						}
						
						d_command_executor.Abort();
						
						d_jog_timer->stop();
						d_control_timer->stop();
					}
				}
			}
			else if(keyword.compare(QStringLiteral("PRBRPT")) == 0)
			{
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("PRBGPA")) == 0)		// search distance, probe mode, max force
			{
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("PRBLPA")) == 0)		// touch speed, probing accuracy, prehit/retract.
			{
				this->Update_Touch_Parameters(arguments);
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("PRBPIN")) == 0)		// PRBPIN X,Y,Z or PRBPIN X,Y,Z,M1,M2,M3,M4,M5,M6,M7,M8,M9,TIP_DIAM,LEN
			{
				p1 = d_tool_offset;
				this->Update_Tool_Data(arguments);
				
				// takes into account the shift in position from a change in the tool offset
				pos = d_command_executor.Current_Position();
				pos += (d_tool_offset - p1);
				
				d_command_executor.Set_Current_Position(pos);
				
				emit Tool_Offset_Changed(d_tool_offset.x,d_tool_offset.y,d_tool_offset.z,d_active_tip_radius * 2.0);
				emit Position_Changed(pos.x,pos.y,pos.z);
				
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("PRBHTY")) == 0)		// PRBHTY PH9,0.0,0.0
			{
				if(arguments.startsWith(QStringLiteral("PH9"),Qt::CaseInsensitive))
				{
					if(Get_Head_Angles(arguments,&a,&b))
					{
						d_head_angle_a = a;
						d_head_angle_b = b;
						
						emit Head_AB_Changed(a,b);
					}
				}
				
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("PRBTYP")) == 0)		// PRBTYP TP6
			{
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("PROBE")) == 0)		// PROBE X,Y,Z,W,I,J,K
			{
				if(d_estop_status)
				{
					this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x14000040 E-Stop Push Button Pressed\r\n");
				}
				else
				{
					if(this->Get_Touch_Data(arguments,&pos,&vec))
					{
						if(this->Is_Position_Inside_Volume(pos,&x_error,&y_error,&z_error))
						{
							p1 = pos + d_prehit_distance * vec + d_active_tip_radius * vec;
							
							command.command_type = TCommandExecutor::COMMAND_MOVE;
							command.target = p1;
							command.speed = d_move_speed;
							command.completion_text = QString();
							
							d_command_executor.Add_Command(command);
							
							p2 = pos + d_active_tip_radius * vec + this->Get_Noise_Offset(vec);
							
							command.command_type = TCommandExecutor::COMMAND_TOUCH;
							command.target = pos;
							command.speed = d_touch_speed;
							command.completion_text = QString("POINT %1, %2, %3, 0.00000, %4, %5, %6\r\n")
									.arg(p2.x,0,'f',5)
									.arg(p2.y,0,'f',5)
									.arg(p2.z,0,'f',5)
									.arg(vec.i,0,'f',5)
									.arg(vec.j,0,'f',5)
									.arg(vec.k,0,'f',5);
							
							d_command_executor.Add_Command(command);
							
							command.command_type = TCommandExecutor::COMMAND_MOVE;
							command.target = p1;
							command.speed = d_move_speed;
							command.completion_text = QString();
							
							d_command_executor.Add_Command(command);
							
							if(!d_control_timer->isActive())
							{
								d_control_timer->start(TIMER_INTERVAL);
							}
						}
						else
						{
							if(x_error)
							{
								emit Log_Text(QStringLiteral("ERROR CE ,S%1 ,TR_CE ,0x1400104d Position out of the machine stroke limit on X axis"));
								this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x1400104d Position out of the machine stroke limit on X axis\r\n");
								d_shift_out_enabled = false;
							}
							
							if(y_error)
							{
								emit Log_Text(QStringLiteral("ERROR CE ,S%1 ,TR_CE ,0x1400204d Position out of the machine stroke limit on Y axis"));
								this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x1400204d Position out of the machine stroke limit on Y axis\r\n");
								d_shift_out_enabled = false;
							}
							
							if(z_error)
							{
								emit Log_Text(QStringLiteral("ERROR CE ,S%1 ,TR_CE ,0x1400304d Position out of the machine stroke limit on Z axis"));
								this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x1400304d Position out of the machine stroke limit on Z axis\r\n");
								d_shift_out_enabled = false;
							}
							
							d_command_executor.Abort();
							
							d_jog_timer->stop();
							d_control_timer->stop();
						}
					}
				}
			}
			else if(keyword.compare(QStringLiteral("READTP")) == 0)
			{
				this->Send("READTP      20.00000\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("REFRCSY")) == 0)
			{
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("RTCTIME")) == 0)
			{
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("SCLTMP")) == 0)		// command may not have trailing parameters
			{
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("SET_PAR")) == 0)
			{
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("SHOW")) == 0)
			{
				if(arguments.compare(QStringLiteral("ACCEL"),Qt::CaseInsensitive) == 0)
				{
					this->Send("ACCEL   1000.00000,  1000.00000,  1000.00000,  1000.00000,  1000.00000\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("GEO"),Qt::CaseInsensitive) == 0)
				{
					if(d_geo_enabled)
					{
						this->Send("GEO TRUE\r\n",send_channel_2);
					}
					else
					{
						this->Send("GEO FALSE\r\n",send_channel_2);
					}
				}
				else if(arguments.compare(QStringLiteral("HAVE_PH9"),Qt::CaseInsensitive) == 0)
				{
					this->Send("HAVE_PH9 1\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("MAXACC"),Qt::CaseInsensitive) == 0)
				{
					this->Send("MAXACC   1000.00000,  1000.00000,  1000.00000,  1000.00000,  1000.00000\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("MAXVEL"),Qt::CaseInsensitive) == 0)
				{
					this->Send("MAXVEL     500.00000,   500.00000,   500.00000,   500.00000,   500.00000\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("MOVPAR"),Qt::CaseInsensitive) == 0)
				{
					write_text = QString("MOVPAR %1, %2, %3, 0.00000, 0.00000, 0.00000, 0.00000, 0.00000\r\n")
					.arg(d_move_speed,0,'f',5)
					.arg(d_move_speed,0,'f',5)
					.arg(d_move_speed,0,'f',5);
					
					this->Send(write_text.toLatin1(),send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("MAXSTROKESW"),Qt::CaseInsensitive) == 0)
				{
					pos = d_max_machine;
					
					write_text = QString("MAXSTROKESW  %1, %2, %3, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000\r\n").arg(pos.x,11,'f',6).arg(pos.y,11,'f',6).arg(pos.z,11,'f',6);
					
					this->Send(write_text.toLatin1(),send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("PRBCONFG"),Qt::CaseInsensitive) == 0)
				{
					this->Send("PRBCONFG    0\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("PRBATT"),Qt::CaseInsensitive) == 0)
				{
					this->Send("PRBATT  0.05000, 0.02500, 0.07500,      10, 1.00000\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("PRBGPA"),Qt::CaseInsensitive) == 0)
				{
					this->Send("PRBGPA  2.50000,      0, 2.50000\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("PRBHTY"),Qt::CaseInsensitive) == 0)
				{
					write_text = QString("PRBHTY  PH9     ,     %1,     %2\r\n").arg(d_head_angle_a,0,'f',5).arg(d_head_angle_b,0,'f',5);
					this->Send(write_text.toLatin1(),send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("PRBLPA"),Qt::CaseInsensitive) == 0)
				{
					write_text = QString("PRBLPA %1,     0, %2\r\n").arg(d_touch_speed,11,'f',5).arg(d_prehit_distance,11,'f',5);
					this->Send(write_text.toLatin1(),send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("PRBPIN"),Qt::CaseInsensitive) == 0)
				{
					write_text = QString("PRBPIN %1, %2, %3,\r\n       0.000000, 0.000000, 0.000000,\r\n       0.000000, 0.000000, 0.000000,\r\n       0.000000, 0.000000, 0.000000,\r\n       %4, -1\r\n%\r\n")
					.arg(d_tool_offset.x,11,'f',5)
					.arg(d_tool_offset.y,11,'f',5)
					.arg(d_tool_offset.z,11,'f',5)
					.arg(d_active_tip_radius * 2.0,0,'f',5);
					
					this->Send(write_text.toLatin1(),send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("PRT_SENS"),Qt::CaseInsensitive) == 0)	// error
				{
					this->Send("PRT_SENS      0\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("MINSTROKESW"),Qt::CaseInsensitive) == 0)
				{
					pos = d_min_machine;
					
					write_text = QString("MINSTROKESW  %1, %2, %3, 0.000000, 0.000000, 0.000000, 0.000000, 0.000000\r\n").arg(pos.x,11,'f',6).arg(pos.y,11,'f',6).arg(pos.z,11,'f',6);
					
					this->Send(write_text.toLatin1(),send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("REF_MARK"),Qt::CaseInsensitive) == 0)	// error
				{
					this->Send("REF_MARK  0.00000, 0.00000, 0.00000, 0.00000, 0.00000, 0.00000\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("RELAYS"),Qt::CaseInsensitive) == 0)
				{
					this->Send("RELAYS TRUE\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("REVRS"),Qt::CaseInsensitive) == 0)	// error
				{
					this->Send("REVRS        -4,   -4,   4,   0,   0,   0\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("SCAPAR"),Qt::CaseInsensitive) == 0)
				{
					this->Send("SCAPAR  0, 2.00000, 5.00000, SCA\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("SCL_RES"),Qt::CaseInsensitive) == 0)	// error
				{
					this->Send("SCL_RES  0.00100, 0.00100, 0.00100, 0.00100, 0.00100, 0.00100\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("SCN_ACC"),Qt::CaseInsensitive) == 0)
				{
					this->Send("SCN_ACC  10.00000\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("SCNP_OFF"),Qt::CaseInsensitive) == 0)
				{
					this->Send("SCNP_OFF  0.10000\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("SP6_MTX"),Qt::CaseInsensitive) == 0)
				{
					this->Send("SP6MTX 0.100000,  0.00000,  0.000000,\r\n       0.000000, 0.100000, 0.000000,\r\n       0.000000, 0.000000, -0.100000\r\n%\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("SRCNUM"),Qt::CaseInsensitive) == 0)
				{
					this->Send("SRCNUM 014a03b2c2603ca2e19bba72104eaf5ed53774900c3bd33a6dfff6f7c324871cd9\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("VOLCOMP"),Qt::CaseInsensitive) == 0)
				{
					this->Send("VOLCOMP 129\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("WKPCMP"),Qt::CaseInsensitive) == 0)
				{
					this->Send("WKPCMP FALSE\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("X_SENS"),Qt::CaseInsensitive) == 0)	// error
				{
					this->Send("X_SENS      0\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("Y_SENS"),Qt::CaseInsensitive) == 0)	// error
				{
					this->Send("Y_SENS      0\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("Z_SENS"),Qt::CaseInsensitive) == 0)	// error
				{
					this->Send("Z_SENS      0\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("VOLCOMPTYPE"),Qt::CaseInsensitive) == 0)
				{
					this->Send("VOLCOMPTYPE 1\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("AUTZER"),Qt::CaseInsensitive) == 0)
				{
					this->Send("AUTZER TRUE\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("DYNCOMP"),Qt::CaseInsensitive) == 0)
				{
					this->Send("DYNCOMP FALSE\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("ESTOP"),Qt::CaseInsensitive) == 0)
				{
					if(d_estop_status)
					{
						write_text = QStringLiteral("ESTOP TRUE\r\n");
					}
					else
					{
						write_text = QStringLiteral("ESTOP FALSE\r\n");
					}
					
					this->Send(write_text.toLatin1(),send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("PRBSTATUS"),Qt::CaseInsensitive) == 0)
				{
					this->Send("PRBSTATUS TOUCH TRIGGER, 0\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("MACHINETYPE"),Qt::CaseInsensitive) == 0)
				{
					this->Send("MACHINETYPE DMP\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("MAPSTAT"),Qt::CaseInsensitive) == 0)
				{
					this->Send("MAPSTAT LOCKED\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("PRBTYP"),Qt::CaseInsensitive) == 0)
				{
					this->Send("PRBTYP TP2\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("TEMPCOMPTYPE"),Qt::CaseInsensitive) == 0)
				{
					this->Send("TEMPCOMPTYPE 0\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("SERIALNUMBER"),Qt::CaseInsensitive) == 0)
				{
					this->Send("SERIALNUMBER 1234\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("SCLTMP"),Qt::CaseInsensitive) == 0)
				{
					this->Send("SCLTMP       20.08060,    20.01810,    19.91893,    20.00000,     0.00000,    40.00000\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("WKPPAR"),Qt::CaseInsensitive) == 0)
				{
					this->Send("WKPPAR      19.9783613032,     0.0000000000,    20.0000000000\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("ERRHISTNUM"),Qt::CaseInsensitive) == 0)
				{
					this->Send("ERRHISTNUM 2\r\n",send_channel_2);
				}
				else if(arguments.startsWith(QStringLiteral("ERRDIAGN"),Qt::CaseInsensitive))
				{
					// SHOW ERRDIAGN, <N>, <TYPE>
					list = arguments.split(',');
					
					if(list.size() > 2)
					{
						if(list[2].compare(QStringLiteral("CURRCOM"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN CURRCOM   NOINPUTCOMMAND\r\n");
						}
						else if(list[2].compare(QStringLiteral("ERRCOD"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN ERRCOD 0x1200001a\r\n");
						}
						else if(list[2].compare(QStringLiteral("ERRTYP"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN ERRTYP    Weak - Motors are not turned off\r\n");
						}
						else if(list[2].compare(QStringLiteral("ERRPOSLOG"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN ERRPOSLOG EPL\r\n");
						}
						else if(list[2].compare(QStringLiteral("ERRNUMPOS"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN ERRNUMPOS 41\r\n");
						}
						else if(list[2].compare(QStringLiteral("ERRPOS"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN ERRPOS 23.582 158.494 196.570 0.000 0.000 0.000 0.000 0.000 0.000\r\n");
						}
						else if(list[2].compare(QStringLiteral("ERRVEL"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN ERRVEL -3.939 -0.695 0.000 0.000 0.000 0.000\r\n");
						}
						else if(list[2].compare(QStringLiteral("INPUTSTR"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN INPUTSTR Some nonsense command.\r\n");
						}
						else if(list[2].compare(QStringLiteral("MODULE"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN MODULE   PI - Protocol Interface (layer 2)\r\n");
						}
						else if(list[2].compare(QStringLiteral("SETTIME"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN SETTIME   11/04/2021 13-55-46.050\r\n");
						}
						else if(list[2].compare(QStringLiteral("RESTIME"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN RESETTIME   11/04/2021 13-56-46.050\r\n");
						}
						else if(list[2].compare(QStringLiteral("LABEL"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN LABEL   Command argument has an invalid value\r\n");
						}
						else if(list[2].compare(QStringLiteral("LOCATION"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN LOCATION   File:../../src/ci.cpp - Function:ciHandlePiErr - Line:598\r\n");
						}
						else if(list[2].compare(QStringLiteral("MANAUTO"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN MANAUTO   1\r\n");
						}
						else if(list[2].compare(QStringLiteral("JOGCNC"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN JOGCNC   1\r\n");
						}
						else if(list[2].compare(QStringLiteral("PRCHNG"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN PRCHNG   0\r\n");
						}
						else if(list[2].compare(QStringLiteral("PRBSTAT"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN PRBSTAT   0\r\n");
						}
						else if(list[2].compare(QStringLiteral("SAFEBARR"),Qt::CaseInsensitive) == 0)
						{
							write_text = QStringLiteral("ERRDIAGN SAFEBARR   0\r\n");
						}
						else
						{
							write_text = QStringLiteral("ERROR    PI      ,S%1     ,TR_PI   ,0x1200001a Command argument has an invalid value\r\n");
						}
						
						this->Send(write_text.toLatin1(),send_channel_2);
					}
				}
				else if(arguments.compare(QStringLiteral("X_NCESCL"),Qt::CaseInsensitive) == 0)
				{
					this->Send("X_NCESCL 0.010\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("Y_NCESCL"),Qt::CaseInsensitive) == 0)
				{
					this->Send("Y_NCESCL 0.010\r\n",send_channel_2);
				}
				else if(arguments.compare(QStringLiteral("Z_NCESCL"),Qt::CaseInsensitive) == 0)
				{
					this->Send("Z_NCESCL 0.010\r\n",send_channel_2);
				}
			}
			else if(keyword.compare(QStringLiteral("VERSION")) == 0)
			{
				this->Send("00-000-000-00000 FDC V16.01.0001 DATE: 01/01/01 TIME: 10:10:10\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("WKPPAR")) == 0)
			{
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("WKPTCSY")) == 0)
			{
				this->Send("%\r\n",send_channel_2);
			}
			else if(keyword.compare(QStringLiteral("WRISTCSY")) == 0)
			{
				this->Send("%\r\n",send_channel_2);
			}
			
#ifdef DEBUG_MODE_LEVEL_0
			else
			{
				qWarning(QString("IGND: %1").arg(line).toLatin1());
			}
#endif
		}
		
	}while(!(rx_line_index < 0));
}

QByteArray TMachineDC::KeyPress_Done(void) const
{
	QByteArray							data("KEYDONE\r\n");
	
	return data;
}

QByteArray TMachineDC::KeyPress_DelPnt(void) const
{
	QByteArray							data("KEYDELPNT\r\n");
	
	return data;
}

void TMachineDC::Set_EStop(
	const bool 							state)
{
	d_estop_status = state;
	
	if(d_estop_status)
	{
		d_command_executor.Abort();
		
		d_jog_timer->stop();
		d_control_timer->stop();
		
		this->Send("ERROR    CE      ,S%1     ,TR_CE   ,0x14000040 E-Stop Push Button Pressed\r\n");
	}
}

bool TMachineDC::Manual_Touch(
	const TVector3						&pos,
	const TVector3						&vec)
{
	TCommandExecutor::TCommand			command;
	QString								text;
	QByteArray							bytes;
	TVector3							touch_pos(pos);
	TVector3							touch_vec(vec);
	TVector3							clear_pos;
	bool								x_error,y_error,z_error;
	static const double					ZERO_EPSILON(0.001);

	if(d_estop_status)
	{
		d_last_error = QStringLiteral("ERR:  Machine in E-Stop");
		return false;
	}
	
	bytes.resize(2);
	
	if(touch_vec.Length() > ZERO_EPSILON)
	{
		touch_vec.Normal();
	}
	else
	{
		d_last_error = QStringLiteral("ERR:  Touch vector not defined");
		return false;
	}
	
	text = QString("Touch %1, %2, %3, %4, %5, %6")
				.arg(touch_pos.x,0,'f',3)
				.arg(touch_pos.y,0,'f',3)
				.arg(touch_pos.z,0,'f',3)
				.arg(touch_vec.i,0,'f',5)
				.arg(touch_vec.j,0,'f',5)
				.arg(touch_vec.k,0,'f',5);
	
	emit Log_Text(text);
	
	clear_pos = touch_pos;
	
	// touch IJK is approach but machine expects surface normal
	touch_vec *= -1.0;

	text = QString("JOGPOINT %1, %2, %3, 0.000000, %4, %5, %6\r\n")
			.arg(touch_pos.x,0,'f',6)
			.arg(touch_pos.y,0,'f',6)
			.arg(touch_pos.z,0,'f',6)
			.arg(touch_vec.i,0,'f',6)
			.arg(touch_vec.j,0,'f',6)
			.arg(touch_vec.k,0,'f',6);
			
	text.insert(0,(unsigned char)(0x4f));
	text.insert(0,(unsigned char)(0x8f));	// prefix is 0x8f4f
	
	if(this->Is_Position_Inside_Volume(clear_pos,&x_error,&y_error,&z_error))
	{
		command.command_type = TCommandExecutor::COMMAND_MOVE;
		command.target = clear_pos;
		command.speed = d_jog_highspeed;
		command.completion_text = text;
		
		d_command_executor.Add_Command(command);
		
		if(!d_control_timer->isActive())
		{
			d_control_timer->start(TIMER_INTERVAL);
		}
	}
	else
	{
		text = QString();
		
		if(x_error)
		{
			text += QStringLiteral("X");
		}
		
		if(y_error)
		{
			text += QStringLiteral("Y");
		}
		
		if(z_error)
		{
			text += QStringLiteral("Z");
		}
		
		d_last_error = QString("ERR:  Axis %1 outside of machine volume.").arg(text);

		return false;
	}

	return true;
}


TVector3 TMachineDC::Get_Move_Data(
	const QString 						&text) const
{
	QStringList							list;
	TVector3							pos;
	
	list = text.split(',');
	
	if(list.size() > 2)
	{
		pos.x = list[0].toDouble();
		pos.y = list[1].toDouble();
		pos.z = list[2].toDouble();
	}
	
	return pos;
}

double TMachineDC::Get_Move_Speed(
	const QString 						&text) const
{
	QStringList							list;
	double								speed(500);
	
	list = text.split(',');
	
	if(list.size() > 2)
	{
		speed = list[0].toDouble();
		speed += list[1].toDouble();
		speed += list[2].toDouble();
		
		speed /= 3.0;
	}
	
	return speed;
}

bool TMachineDC::Move_To(
	const TVector3						&pos)
{
	TCommandExecutor::TCommand			command;
	TVector3							target(pos);
	
	if(d_estop_status)
	{
		d_last_error = QStringLiteral("ERR:  Machine in E-Stop");
		return false;
	}

	target -= d_tool_offset;

	if(target.x < d_min_machine.x) target.x = d_min_machine.x;
	if(target.y < d_min_machine.y) target.y = d_min_machine.y;
	if(target.z < d_min_machine.z) target.z = d_min_machine.z;
	if(target.x > d_max_machine.x) target.x = d_max_machine.x;
	if(target.y > d_max_machine.y) target.y = d_max_machine.y;
	if(target.z > d_max_machine.z) target.z = d_max_machine.z;
	
	target += d_tool_offset;
	
	command.command_type = TCommandExecutor::COMMAND_MOVE;
	command.target = pos;
	
	command.speed = d_move_speed;
	
	command.completion_text = QString();
	
	d_command_executor.Add_Command(command);
	
	if(!d_jog_timer->isActive())
	{
		d_jog_timer->start(TIMER_INTERVAL);
	}
	
	return true;
}

bool TMachineDC::Get_Touch_Data(
	const QString						&text,
	TVector3							* const xyz,
	TVector3							* const ijk) const
{
	QStringList							list;
	TVector3							vec;
	static const double					ZERO_EPSILON(0.001);
	
	assert(xyz);
	assert(ijk);
	
	list = text.split(',');
	
	if(list.size() > 6)
	{
		(*xyz).x = list[0].toDouble();
		(*xyz).y = list[1].toDouble();
		(*xyz).z = list[2].toDouble();
		
		vec.i = list[4].toDouble();
		vec.j = list[5].toDouble();
		vec.k = list[6].toDouble();
		
		if(vec.Length() > ZERO_EPSILON)
		{
			vec.Normal();
		}
		else
		{
			vec.Set(0,0,1);
		}
		
		*ijk = vec;
		
		return true;
	}
	
	return false;
}

bool TMachineDC::Jog_Move(
	const TVector3						&direction)
{
	TCommandExecutor::TCommand			command;
	TVector3							target(d_command_executor.Current_Position());
	
	if(d_estop_status)
	{
		d_last_error = QStringLiteral("ERR:  Machine in E-Stop");
		return false;
	}
	
	if(d_control_timer->isActive())
	{
		d_last_error = QStringLiteral("ERR:  Machine currently active");
		return false;
	}
	
	target += (d_jog_distance * direction);

	target -= d_tool_offset;
	
	if(target.x < d_min_machine.x) target.x = d_min_machine.x;
	if(target.y < d_min_machine.y) target.y = d_min_machine.y;
	if(target.z < d_min_machine.z) target.z = d_min_machine.z;
	if(target.x > d_max_machine.x) target.x = d_max_machine.x;
	if(target.y > d_max_machine.y) target.y = d_max_machine.y;
	if(target.z > d_max_machine.z) target.z = d_max_machine.z;
	
	target += d_tool_offset;
	
	command.command_type = TCommandExecutor::COMMAND_MOVE;
	command.target = target;
	
	command.speed = d_move_speed;
	
	command.completion_text = QString();
	
	d_command_executor.Add_Command(command);
	
	if(!d_jog_timer->isActive())
	{
		d_jog_timer->start(TIMER_INTERVAL);
	}
	
	return true;
}

bool TMachineDC::Get_Head_Angles(
	const QString						&text,
	double								* const a,
	double								* const b) const
{
	QStringList							list;
	
	assert(a);
	assert(b);
	
	list = text.split(',');		// PH9, A, B
	
	if(list.size() > 2)
	{
		// list[0] is PH9
		*a = list[1].toDouble();
		*b = list[2].toDouble();
		
		return true;
	}
	
	return false;
}


void TMachineDC::Update_Touch_Parameters(
	const QString						&text)
{
	QStringList							list;
	
	list = text.split(',');
	
	if(list.size() > 2)
	{
		d_touch_speed = list[0].toDouble();
		d_prehit_distance = list[2].toDouble();
	}
}

void TMachineDC::Update_Tool_Data(
	const QString						&text)
{
	QStringList							list;
	TVector3							pos;
	
	list = text.split(',');
	
	if(list.size() > 2)
	{
		pos.x = list[0].toDouble();
		pos.y = list[1].toDouble();
		pos.z = list[2].toDouble();
		
		d_tool_offset = pos;
	}
	
	if(list.size() > 12)
	{
		d_prbpin_tip_radius = list[12].toDouble() / 2.0;
	}
	
	if(d_use_override_tip_radius)
	{
		d_active_tip_radius = d_override_tip_radius;
	}
	else
	{
		d_active_tip_radius = d_prbpin_tip_radius;
	}
}
